@@ -102,25 +102,25 @@ class SoftwaresExtension(Extension):
                 "terms": {
                     "field": "software.keyword",
                     "size": 500,
-                    "order": {"_key": "asc"}
+                    "order": {"_count": "desc"}
                 },
-                "aggs": {
-                    "casinos": {
-                        "top_hits": {
-                            "sort": [
-                                {
-                                    "rating": {
-                                        "order": "desc"
-                                    }
-                                }
-                            ],
-                            "_source": {
-                                "includes": ["title", "path", "alt_title", "rating"]
-                            },
-                            "size": 3
-                        }
-                    }                
-                }
+                # "aggs": {
+                #     "casinos": {
+                #         "top_hits": {
+                #             "sort": [
+                #                 {
+                #                     "rating": {
+                #                         "order": "desc"
+                #                     }
+                #                 }
+                #             ],
+                #             "_source": {
+                #                 "includes": ["title", "path", "alt_title", "rating"]
+                #             },
+                #             "size": 3
+                #         }
+                #     }                
+                # }
             }
         }
 
@@ -141,7 +141,7 @@ class SoftwaresExtension(Extension):
             return ''
 
         t = current_app.jinja_env.get_template('_table-simple.html')
-        s = t.render(pages=pages, items=resp['aggregations']['software']['buckets'])
+        s = t.render(pages=pages, title='Software', items=resp['aggregations']['software']['buckets'])
 
         return s
 
@@ -153,6 +153,7 @@ class SoftwaresExtension(Extension):
 
         return CallBlock(self.call_method("_render", args), [], [], body).set_lineno(lineno)
 
+
 class CountriesExtension(Extension):
     tags = set(['countries'])
 
@@ -225,25 +226,25 @@ class BankingExtension(Extension):
                 "terms": {
                     "field": "deposits.keyword",
                     "size": 500,
-                    "order": {"_key": "asc"}
+                    "order": {"_count": "desc"}
                 },
-                "aggs": {
-                    "casinos": {
-                        "top_hits": {
-                            "sort": [
-                                {
-                                    "rating": {
-                                        "order": "desc"
-                                    }
-                                }
-                            ],
-                            "_source": {
-                                "includes": ["title", "path", "alt_title", "rating"]
-                            },
-                            "size": 3
-                        }
-                    }                
-                }
+                # "aggs": {
+                #     "casinos": {
+                #         "top_hits": {
+                #             "sort": [
+                #                 {
+                #                     "rating": {
+                #                         "order": "desc"
+                #                     }
+                #                 }
+                #             ],
+                #             "_source": {
+                #                 "includes": ["title", "path", "alt_title", "rating"]
+                #             },
+                #             "size": 3
+                #         }
+                #     }                
+                # }
             }
         }
 
@@ -264,7 +265,7 @@ class BankingExtension(Extension):
             return ''
 
         t = current_app.jinja_env.get_template('_table-simple.html')
-        s = t.render(pages=pages, items=resp['aggregations']['deposits']['buckets'])
+        s = t.render(pages=pages, title='Payment Method', items=resp['aggregations']['deposits']['buckets'])
 
         return s
 
@@ -284,17 +285,20 @@ class RatingExtension(Extension):
         super(RatingExtension, self).__init__(environment)
 
     def _render(self, args, caller=None):
-        logger.info('Args: {0}'.format(args))
+        logger.info(f'Casinos Args IN: {args}')
 
         default = ['casino', True, 10]
-        args = args[:2] + default[len(args):] + args[2:]
+        _ca = len(default)
+        args = args[:_ca] + default[len(args):] + args[_ca:]
+
+        logger.info('Casinos Args OUT: {0}'.format(args))
 
         service = args[0]
         has_filters = args[1]
         count = int(args[2])
 
         tags = args[3:]
-        logger.info('Tags: {0}'.format(tags))
+        logger.info('Casinos Tags: {0}'.format(tags))
 
         _afields = ['software', 'licences', 'deposits', 'withdrawal', 'games']
         _aggs = {
@@ -350,7 +354,7 @@ class RatingExtension(Extension):
 
         deposits_primary = ['Visa', 'Trustly', 'PayPal', 'Skrill', 'Neteller']
 
-        t = current_app.jinja_env.get_template('_rating.html')
+        t = current_app.jinja_env.get_template('_ext-rating_casinos.html')
         s = t.render(
             pages=pages, 
             has_filters=has_filters, 
@@ -374,38 +378,6 @@ class RatingExtension(Extension):
         return CallBlock(self.call_method("_render", args=[List(args)]), [], [], []).set_lineno(lineno)          
 
 
-# class SlotsExtension(Extension):
-#     tags = set(['slots'])
-
-#     def __init__(self, environment):
-#         super(SlotsExtension, self).__init__(environment)
-
-#     def _render(self, caller):
-#         slots, found = Page.get(
-#             category='slot', 
-#             is_active=True, 
-#             is_searchable=True, 
-#             is_redirect=False, 
-#             locale=g.language, 
-#             _source=['alias', 'title', 'cover', 'software'], 
-#             _count=500,
-#             _sort=[
-#                 {'title.keyword': {'order': 'asc'}}
-#             ]
-#         )
-#         t = current_app.jinja_env.get_template('_slots.html')
-#         s = t.render(slots=slots, found=found)
-#         return s
-
-#     def parse(self, parser):
-#         lineno = next(parser.stream).lineno
-
-#         body = ''
-#         args = []
-
-#         return CallBlock(self.call_method("_render", args), [], [], body).set_lineno(lineno)         
-
-
 class TopSlotsExtension(Extension):
     tags = set(['top_slots'])
 
@@ -639,14 +611,14 @@ class TopAlexaExtension(Extension):
             is_searchable=True,
             is_redirect=False,  
             locale=g.language, 
-            _source=['title', 'path', 'alt_title', 'rank_alexa'],
+            _source=['title', 'path', 'alt_title', 'rank_alexa', 'geo'],
             _count=30,
             _sort=[
                 {'rank_alexa': {'order': 'asc'}}
             ]
         )
    
-        t = current_app.jinja_env.get_template('_top-alexa.html')
+        t = current_app.jinja_env.get_template('_ext-top_alexa.html')
         s = t.render(pages=pages)
         return s
 
@@ -751,9 +723,9 @@ class SlotsExtension(Extension):
         super(SlotsExtension, self).__init__(environment)
 
     def _render(self, args, caller=None):
-        logger.info('Slots Args IN: {0}'.format(args))
+        logger.info(f'Slots Args IN: {args}')
 
-        default = [10, 'slots_top', False] # cnt, template, has_filters, [tags...]
+        default = [10, 'rating_slots', False] # cnt, template, has_filters, [tags...]
         _ca = len(default)
         args = args[:_ca] + default[len(args):] + args[_ca:]
 
@@ -766,29 +738,44 @@ class SlotsExtension(Extension):
         tags = args[3:]
         logger.info('Slots Tags: {0}'.format(tags))
 
-        if has_filters:
-            pass
-        else:
-            slots, found = Page.get(
-                category='slot', 
-                is_active=True, 
-                is_searchable=True, 
-                is_redirect=False, 
-                locale=g.language, 
-                # locale_available=g.language,
-                _source=['alias', 'title', 'cover', 'software'], 
-                _count=500,
-                _sort=[
-                    {'title.keyword': {'order': 'asc'}}
-                ]
-            )            
+        _afields = ['software', 'volatility', 'themes', 'slot_features']
+        _aggs = {
+            item: {
+                "terms": {
+                    "field": "{0}.keyword".format(item),
+                    "size": 500,
+                    "order": {"_key": "asc"}
+                }
+            } for item in _afields
+        }
 
+        slots, found, aggs, id = Page.slots_by_context(
+            is_searchable=True, 
+            is_redirect=False, 
+            _locale=g.language, 
+            _source=[
+                'alias', 
+                'title', 
+                'cover', 
+                'software'
+            ] + _afields, 
+            _count=count,
+            _aggs=_aggs,
+        )            
+
+        if id and has_filters:
+            _c = dumps(aggs)
+            current_app.redis.hset('aggs', id, _c)
+            logger.info('Cache aggs {0}: {1}'.format(id, len(_c)))
+            
         t = current_app.jinja_env.get_template('_ext-{0}.html'.format(template))
         s = t.render(
             slots=slots, 
-            found=found, 
             has_filters=has_filters, 
+            found=found, 
+            id=id, 
             count=count,
+            tags=tags,
         )
         return s
 
@@ -932,6 +919,9 @@ class ComparatorExtension(Extension):
 
         try:
             casinos = []
+            casinos = []
+            select_enable = False
+
             if request.cookies.get('comparator'):
                 cmp_list = loads(unquote(request.cookies.get('comparator')))
                 logger.info(f'Casinos for compare: {cmp_list}')
@@ -946,11 +936,40 @@ class ComparatorExtension(Extension):
                     _count=3,
                 )
 
+            if len(casinos) <= 3:
+                recommended, _ = Page.provider_by_context(
+                    country=current_user.country_full,
+                    is_searchable=True,
+                    is_redirect=False,
+                    _locale=g.language,
+                    _source=[
+                        'alias', 
+                        'title', 
+                        'theme_color', 
+                        'logo', 
+                    ], 
+                    _exclude=[casino._id for casino in casinos],
+                    _count=6
+                )
+
+                def process_casino(i):
+                    obj = Page({
+                        'title': 'Casino ' + str(i + 1),
+                        'logo': '/static/media/images/default-cover.png',
+                        'theme_color': '#1b114d'
+                    }, Page.get_urlsafe_string(6))
+                    return obj
+
+                if len(casinos) < 3:
+                    select_enable = True
+
+                casinos = casinos + [process_casino(i) for i in range(0, 3 - len(casinos))]
+
             t = current_app.jinja_env.get_template('_ext-comparator.html')
             attrs = {
                 'General': [
-                    'logo',
-                    'path',
+                    'logo', 
+                    'path', 
                     'establishedon',
                     'website',
                     'languages',
@@ -1035,7 +1054,14 @@ class ComparatorExtension(Extension):
                 'welcome_package_wager': 'Wager',
             }
 
-            s = t.render(casinos=casinos, attrs=attrs, names=names, currency_geo=g.currency_geo)
+            s = t.render(
+                casinos=casinos, 
+                attrs=attrs, 
+                names=names, 
+                currency_geo=g.currency_geo,
+                select_enable=select_enable,
+                recommended=recommended
+            )
             return s
         except Exception as e:
             logger.error(f'Comparator Exception: {e}')
@@ -1060,7 +1086,7 @@ class BlogExtension(Extension):
         super(BlogExtension, self).__init__(environment)
 
     def _render(self, args, caller=None):
-        logger.info('Args: {0}'.format(args))
+        logger.info('blog Args: {0}'.format(args))
 
         default = [12, 3, True]
         _ca = len(default)
@@ -1123,6 +1149,8 @@ class TopPromoExtension(Extension):
         super(TopPromoExtension, self).__init__(environment)
 
     def _render(self, caller=None):
+        logger.info(f'Generate (top_promo) for {current_user.country_full}')
+
         casinos, found = Page.provider_by_context(
             country=current_user.country_full,
             is_searchable=True,
@@ -1136,7 +1164,7 @@ class TopPromoExtension(Extension):
             ], 
             _count=3
         )    
-        t = current_app.jinja_env.get_template('_exp-promo.html')
+        t = current_app.jinja_env.get_template('_ext-promo.html')
         s = t.render(casinos=casinos, found=found)
         return s
 
@@ -1146,3 +1174,126 @@ class TopPromoExtension(Extension):
         body = ''
         args = []
         return CallBlock(self.call_method("_render", args), [], [], body).set_lineno(lineno)
+
+
+class TabMenuExtension(Extension):
+    tags = set(['tab_menu'])
+
+    def __init__(self, environment):
+        super(TabMenuExtension, self).__init__(environment)
+
+    def _render(self, args, caller=None):
+        logger.info('tab_menu Args: {0}'.format(args))
+
+        path = str(args[0])
+
+        _aggs = {
+            item: {
+                "terms": {
+                    "field": "{0}.keyword".format(item),
+                    "size": 500,
+                    "order": {"_key": "asc"}
+                }
+            } for item in ['provider_tags', 'deposits', 'games']
+        }
+
+        _, _, aggs, _ = Page.provider_by_context(
+            country=current_user.country_full,
+            is_searchable=True,
+            is_redirect=False,
+            services=['casino'],
+            _locale=g.language,
+            _count=0,
+            _aggs=_aggs
+        )    
+
+        def process_count(_items, _key):
+            for item in _items:
+                if item['item'] == _key:
+                    return item['count']
+            return None
+
+        pages = [
+            {'path': '/casino/casino-knowledge-base/essential-guide-to-live-dealer-online-casino.html', 'title': 'Live Casinos', 'count': process_count(aggs['games'], 'Live Games')},
+            {'path': '/casino/established-online-casinos/', 'title': 'Big Brand Casinos', 'count': process_count(aggs['provider_tags'], 'Big')},
+            {'path': '/collections/25-licensed-casinos-with-fastest-withdrawals.html', 'title': 'Fast Payout Casinos', 'count': process_count(aggs['provider_tags'], 'Sponsored')},
+            {'path': '/casino/high-traffic-casinos/', 'title': 'High Traffic Casinos', 'count': 30},
+            {'path': '/casino/casino-payment-methods/bitcoin-online-casinos.html', 'title': 'Bitcoin Casinos', 'count': process_count(aggs['deposits'], 'Bitcoin')},
+            {'path': '/casino/casino-payment-methods/casinos-that-use-paypal.html', 'title': 'Paypal Casinos', 'count': process_count(aggs['deposits'], 'PayPal')},
+        ]
+
+        t = current_app.jinja_env.get_template('_ext-tab_menu.html')
+        s = t.render(pages=pages, path=path)
+        return s
+
+    def parse(self, parser):
+        lineno = next(parser.stream).lineno
+
+        args = []
+        while parser.stream.current.type != 'block_end':
+            args.append(parser.parse_expression())
+            parser.stream.skip_if('comma')
+
+        return CallBlock(self.call_method("_render", args=[List(args)]), [], [], []).set_lineno(lineno)          
+
+
+class RelatedPagesExtension(Extension):
+    tags = set(['related_pages'])
+
+    def __init__(self, environment):
+        super(RelatedPagesExtension, self).__init__(environment)
+
+    def _render(self, args, caller=None):
+        logger.info('related_pages Args: {0}'.format(args))
+
+        pages, _ = Page.get(
+            is_active=True, 
+            is_searchable=True, 
+            is_redirect=False, 
+            tags=args[0],
+            locale=g.language, 
+            _source=['path', 'title'], 
+            _count=10,
+            _exclude=[args[1]],
+            _sort=[
+                {'publishedon': {'order': 'desc'}}
+            ]
+        )
+
+        t = current_app.jinja_env.get_template('_ext-related_pages.html')
+        s = t.render(pages=pages)
+        return s
+
+    def parse(self, parser):
+        lineno = next(parser.stream).lineno
+
+        args = []
+        while parser.stream.current.type != 'block_end':
+            args.append(parser.parse_expression())
+            parser.stream.skip_if('comma')
+
+        return CallBlock(self.call_method("_render", args=[List(args)]), [], [], []).set_lineno(lineno)          
+
+
+class ComplaintAddExtension(Extension):
+    tags = set(['complaint_add'])
+
+    def __init__(self, environment):
+        super(ComplaintAddExtension, self).__init__(environment)
+
+    def _render(self, args, caller=None):
+        logger.info('complaint_add Args: {0}'.format(args))
+
+        t = current_app.jinja_env.get_template('_ext-complaint_add.html')
+        s = t.render()
+        return s
+
+    def parse(self, parser):
+        lineno = next(parser.stream).lineno
+
+        args = []
+        while parser.stream.current.type != 'block_end':
+            args.append(parser.parse_expression())
+            parser.stream.skip_if('comma')
+
+        return CallBlock(self.call_method("_render", args=[List(args)]), [], [], []).set_lineno(lineno)          
