@@ -20,7 +20,7 @@ import shortuuid
 from pydantic import BaseModel, validator, EmailStr, conint, conlist, HttpUrl
 from typing import Optional, List
 from cryptography.fernet import Fernet
-
+from typing import Union
 
 class UndefinedEntityAttribute(Exception):
     pass
@@ -741,6 +741,7 @@ class Page(ElasticEntity):
                 'themes:select:themes',
                 'slot_features:select:slot_features',
                 'is_jackpot:bool',
+                'licences_info.number:str',
             ],
             'collection_sort': ['Func_1', 'Func_2', 'Rate', 'Rank', 'Views', 'Activity', 'Alexa', 'Adding', 'AZ', 'Launched', 'RTP', 'MaxWin', 'Jackpot', 'Min.Deposit'],
             'collection_sort_direction': ['Desc', 'Asc'],
@@ -956,6 +957,113 @@ class Page(ElasticEntity):
             return cls.process(resp['hits']['hits']), resp['hits']['total']['value']
 
 
+    @classmethod
+    def slots_by_context(cls, _exclude=[], _count=12, _page=1, _source=["title", "alias", "cover"], _aggs={}, _sorting=None, _locale='en', **params):
+        filter_list =  [ 
+            {"term": {"is_active": True}},
+            {"term": {"category.keyword": 'slot'}},
+        ]
+        if _locale:
+            filter_list.append({"term": {"locale.keyword": _locale}})            
+        for key, value in params.items():
+            if key not in ['sorting', 'cpp', 'query']:
+                if isinstance(value, bool):
+                    filter_list.append({"term": {key: value}})
+                elif isinstance(value, list):
+                    if value:
+                        filter_list.append({"terms": {
+                            "{0}.keyword".format(key): value,
+                        }})
+                else:                
+                    filter_list.append({"term": {"{0}.keyword".format(key): value}})
+            if key == 'query':
+                if value:
+                    filter_list.append(
+                        {"multi_match" : {
+                            "query": value, 
+                            "type": "bool_prefix",
+                            "fields": [
+                                "suggest",
+                                "suggest._2gram",
+                                "suggest._3gram"
+                            ] 
+                        }})
+
+        def get_func():
+            flist = list()
+
+            flist.append({
+                "field_value_factor": {
+                    "field": "rating",
+                    "factor": 1,
+                    "missing": 0
+                },
+                "weight": 2
+            })
+
+
+            flist.append({
+                "filter": {
+                    "term": {
+                        "is_features": True
+                    }
+                },
+                "weight": 3
+            })
+
+            return flist
+
+        qbool = {
+            "must": filter_list,
+            "must_not": {'ids': {'values': _exclude}}
+        }
+
+        q = {
+            "query": {
+                "function_score": {
+                    "query": {
+                        "bool": qbool
+                    },
+                    "boost": 1,
+                    "boost_mode": "multiply",
+                    "functions": get_func()
+                }
+            },
+            "from": (_page - 1)*_count,
+            "size": _count,
+            "_source": _source,
+        }
+
+        _hash = None
+        if _aggs:
+            q["aggs"] = _aggs
+            _hash = hashlib.md5(dumps(q).encode()).hexdigest()
+
+        if _sorting:
+            q["sort"] = _sorting
+
+        resp = current_app.es.search(index=cls._table(), body=q, request_timeout=60, ignore=[400, 404])
+        # logger.debug(u'RESPONSE: {0}'.format(resp))
+
+        if 'status' in resp and resp['status'] in [400, 404]:
+            if _aggs:
+                return [], 0, {}, _hash
+            else:
+                return [], 0
+
+        def process_aggs(_a):
+            _res = {}
+            for k, v in _a.items():
+                _res[k] = [{'item': i['key'], 'count': i['doc_count']} for i in v['buckets']]
+            return _res
+
+        logger.debug(u'RESPONSE slots_by_context took: {0}'.format(resp['took']))
+        if _aggs:
+            return cls.process(resp['hits']['hits']), resp['hits']['total']['value'], process_aggs(resp['aggregations']), _hash
+        else:
+            return cls.process(resp['hits']['hits']), resp['hits']['total']['value']
+
+
 class Guest(AnonymousUserMixin):
     @property
     def is_admin(self):
@@ -1061,17 +1169,30 @@ class Activity(ElasticEntity):
         'name': str,
         'casino': str,
 
+        # ticket
         'subject': str,
         'contacts': str,
         'message': str,
 
+        # subscribe
         'email': str,
 
+        # vote, rate + old review
         'rate': int,
         'casino_id': str,
         'pros': str,
         'cons': str,
 
+        # complaint
+        'username': str,
+        'amount': str,
+        'currency': str,
+        'completedon': DateTimeField,
+        'status': str,
+        'replies': list,
+        'is_active': bool,
+
+        # click
         'url': str,
         'landing': str,
     }
@@ -1665,9 +1786,54 @@ class ActivityFeedbackModel(BaseModel):
     rate: conint(ge=1, le=5)
     casino: str = ...
     casino_id: str = ...
-    name: Optional[str]
-    pros: Optional[str]
-    cons: Optional[str]                
+    # name: Optional[str]
+    # pros: Optional[str]
+    # cons: Optional[str]                
+
+
+class ActivityComplaintModel(BaseModel):
+    name: str = ...
+    country: str = ...
+    casino_selected: dict = ...
+    amount: str = ...
+    currency: str = ...
+    message: str = ...
+    username: str = ...
+    email: EmailStr
+
+
+class ComplaintModel(BaseModel):
+    @validator('*', pre=True, allow_reuse=True)
+    def empty_str_to_none(cls, v):
+        if v == '':
+            return None
+        return v
+
+    name: str = ...
+    username: str = ...
+    email: EmailStr
+    country: str = ...
+    casino_selected: dict = ...
+    amount: str = ...
+    currency: str = ...
+    message: str = ...
+    status: str = ...
+    ip: str = ...
+    is_active: bool = False
+    rate: Optional[conint(ge=1, le=5)] = None
+    replies: List[dict] = []
+    createdon: Union[str, date] = ...
+    completedon: Optional[Union[str, date]] = ...
+
+
+class ComplaintReplyModel(BaseModel):
+    @validator('*', pre=True, allow_reuse=True)
+    def empty_str_to_none(cls, v):
+        if v == '':
+            return None
+        return v
+
+    reply: Optional[str] = ...
 
 
 class TelegramBotModel(BaseModel):
